from unittest import TestCase

from testutils import pub_A, priv_A, pub_B, priv_B, create_fetch_sender
from py_it_crypto.user.user import UserManagement

goDecryptB = '{"protected":"eyJhbGciOiJFQ0RILUVTK0EyNTZLVyIsImVuYyI6IkEyNTZHQ00iLCJlcGsiOnsia3R5IjoiRUMiLCJjcnYiOiJQLTI1NiIsIngiOiJpeHdsY24zdFFQakJnY2JoU0JmMWJHaE5ZalV0ckJwZTNzdkNaUGJsTV9BIiwieSI6Im1HdWVPaTBwQUc1aThLME5VTmdHT2xhTm1HeFFtQVFDMmxNYktoS1lUbmcifSwic2hhcmVkSGVhZGVyIjp7InBheWxvYWQiOiJleUp6YUdGeVpVbGtJam9pTldObVl6STNNV0l0WkdJeE1TMDBOREEzTFRnME56TXRNRGxqTVRFME5ERXlZemRpSWl3aWIzZHVaWElpT2lKeVpXTmxhWFpsY2lJc0luSmxZMlZwZG1WeWN5STZXeUp5WldObGFYWmxjaUpkZlEiLCJwcm90ZWN0ZWQiOiJleUpoYkdjaU9pSkZVekkxTmlKOSIsInNpZ25hdHVyZSI6IlBGOTVwOHMwQVNOYkVSS3hncExSWE5WUUdHb1Jwak92UXIzQ3RZdmN0aUhDYmZOcmM1RmQ5SFpHcmFiczRrWWdhTWtFOG0zTDVLakotRGNxNnFUVnJ3In19","encrypted_key":"tGgz8UpMQiKGer3vJGh8fLnWvx9nYL2ZGJjR-D6Peo6Oz7EkVG53mQ","iv":"0AMTbgJrnoGYuLaA","ciphertext":"hCQT-2dat2PDRBHxw9xsMPf_X5HW-quewB7fawipIUbr7lHNnEMiMyDxiEvmR_sJe_MtiEv7pYg1fAgMmwHh8yGWv23fyQG7g-g1cSC2vG5ArgSDloe4EcOoMkvGxv1hSq3vMI_UNf0VQyqQu5JbeZTWPH0wVAsyVckkEf-5wpS7SH1h8GYGtsLpgF7FhjigBJSvfqwp-yYOiUkqclhYTebIz1gnlcFTHCioX9VwQkeVt5OLqfXi5dgyx1YPHTDNowmr_d9J_-Obn-k4JGPvBXaazIYtfgzBO8b23srNMSsFJhhtE6dADhPHWPZO94Swg4XJNhz884raNCw3J1-p5odlhAsp6Kw4Q7a4VZvVNX2u0k-JNrN4YRqGn_DB1noHnWfAkf6-0-6Jqi_9Sewdj48avGly0M1NRAwigoN7gpMCHepaNbdDiJqHSEEFh10H55ZxX_tgUtdHwrBj89otPtU8mveTDfR7u-JwugwzD2DV76vc96GW_pTLWP9QDOhA9ACkqCVlHPyefSWFNOhaT0IOK7SUFtYjZuzfUVP_eI6SVZBXGk_PEo64puHfhfbhwfmBv1A22BITmkIW1FzSDAI6T05ZryEAyfYjVOd-WXzIUjy2MqRBN-Y6EwuUui7fUKHx02hZek1CnZmo2lbAUk9zMJOMLcmhyVryWsVy8eftIqThHK5X8QNnB7ojlv1Dq_hLQxp0SGIR-q_ga9eqvj_5oU63_s2Pane-yBDkavzUZH9o4PgYOcrm4ouAs1pTuYeqdLMpJVQLckseXxz8A5sIw7SJ5dRbI_wZ3gN16ODTl5U8bxEvPGN4f4BJiXYRiakpwogBESVdECEQIYYUduPxSOQKkOZaBr6hCx4JtbERgHJj5vs_5TsI10LJt_xByvKkslzfGyvvYXspbdQLUsXCTEUntPygdAseDOtro3mwYxf8m9KGcSR1zdy8BSHa8NRqoZMn1Xk","tag":"-FqAT10fjTSxJupw6Wy3xw"}';

goDecryptAB = '{"protected":"eyJlbmMiOiJBMjU2R0NNIiwic2hhcmVkSGVhZGVyIjp7InBheWxvYWQiOiJleUp6YUdGeVpVbGtJam9pWTJFek16SmpNVGd0TnpCbE9DMDBNVFl4TFRrNFpqWXRNMkl4TjJFeE5EQTNPRFl4SWl3aWIzZHVaWElpT2lKeVpXTmxhWFpsY2lJc0luSmxZMlZwZG1WeWN5STZXeUp5WldObGFYWmxjaUlzSW5ObGJtUmxjaUpkZlEiLCJwcm90ZWN0ZWQiOiJleUpoYkdjaU9pSkZVekkxTmlKOSIsInNpZ25hdHVyZSI6ImphVUZ2cWxaVmtwSnp5Z0xkS3JxNHM5VEJSZE5DQUlHSjNBcjRzc0dpVm5QTklsWG1OZ0JqQXZGaExydmptUV9wS3hCbmszZ05FUkpULVQzdDBLX0ZnIn19","recipients":[{"header":{"alg":"ECDH-ES+A256KW","epk":{"kty":"EC","crv":"P-256","x":"aD-XXwe5BVZOnsB8G4u0Or_E3_bxR0TSWORAKezaRwo","y":"x7nhjxUmlbSK2N6uBOKKWDRYiBSbKtfUjIu_rscoDOA"}},"encrypted_key":"OpRu1mbwqXvXlKtl7m7pAKaoNlN20ow73mBPorqSgf50rsASSJsaFA"},{"header":{"alg":"ECDH-ES+A256KW","epk":{"kty":"EC","crv":"P-256","x":"wl1tneiCrj6gMq-1cAZ1WvxWCO8YG9uUpGlLAEKn72Q","y":"zB5LxIZqK5FNX8Tem8YCr93nRojd7kKeGWIaqlmBL7g"}},"encrypted_key":"fWBeM0RERw6wtZEfCPvx80ytQlmVQHCuxYfXhfbolpyvQHopzu3csQ"}],"encrypted_key":"OpRu1mbwqXvXlKtl7m7pAKaoNlN20ow73mBPorqSgf50rsASSJsaFA","iv":"YcKsOU0QWozaTuei","ciphertext":"Py8w6a8K4yr-4RbDprAuB8vNw2Cn86MChIlc6erc0TdN_9ceFaSYcgv9iDDd1rhJV-_8JMzGC6v5sDPsdpk3vQzp8gpJ4CBKFsJzQ-07NnkU5xgWCl-PEZsDxN6mAOEhx0beWMPqqH9U4nblOsX5RxeTIvzMkp7WvAMDk9TQqynns0-PsyEkKl5OWo3JZ3ziKqCKaBEPAj_KTSuJpB6xLGWdJnEq8XUhV4ZtdhY0IxA0ASmEz6FyBcMtU_4X8OR3rZNzHwW2SquVfil0OIts_blmdITw15CQwy8o-4xD7myierIVjQsBy3DarR8eh8ztGLg3YrWuGS5WFdVQLVtYPRLpgnvS48lJXGWDu2Ih843x89PYixljkZ7j3VtXx-ATO9DBby84ty8bm7gHctnvzBXgWsX3ICa_Sd4w9nvmynLewyOJt6JGhYtpGgMOYZDw9gpfwcqixynoOalt4_f-LUq9jB0juPawe60fXJeg-FlyiKrwmcqsbZaggqru_7Y7RCRfeikv0G3mjoyciw1xQSTkJxEr0QjLAOv7K3VZw-Hf19O1R7xSUFfacmXF5zwmIRGtY0HgSOl0BfU9W6mbNPF2WkKG1dWjY2FoA8M1N3ssn8YHqD15R97yOUPLpMWjjMi9mDRMyyA0S8oV9R8YF9Ab3eCKLwWQuG2plcQeWILom1lFLC_HYAyF1cuK1NNuOpq8XeCbQh1MwFmeM3QRhLx2ZjdcvqeMWixtFdz6VaSCFU5nFrgd3ij9fi8dqjSyY6uhq-_yXu57ZJLN1nD7BMLQKZkRRB14AV4Q3MmNKAzVRBobVbCgXIfU394icOH6FUhn2f-f3IJsQi0vZMBToy50URH10amuCaYrx9YoW7rE4nCJwMVEqZPLwsi3jtywcdf7KWmrpH-ZvX9GI1w6NrsqliLMmGEawuVKfGu6_UNv6wSY19MFlsBcWH1lzfg3eb10N4ccZq_fNQ","tag":"t2On3Qv0DM6eERomt0Q9ew"}';

pythonDecryptB = '{"ciphertext":"NIFn5k9zrVpD7iRRTrbPBU0fC1k33rBVHm_fafCHyuOEkB9aiJF50Y318U-oKFBZ49Ufpu5vfTrwWbNj_jmcShBkdYXNYg8-ehVqku_I0LAaa3EBinf7K81cFGvRmHfePCbQLw3ku2SByIuFn0lpf5-DIkY8vJarThbuB4Dlh0qfEfTLoPSQo5_sU8T03CogxYfCD5PGjRe-qbNXRefg831YAJAHjavt8KndRxALGSGqhwR0EXeH4SGexazaSFJRNW15v3cGNg-Pw8pSI_r-cMO9Ve-gzS8kMpjz8Voo-GVkRUSyz3GvsG76GXXuUfwUMVjmFTGNl_M6cJXEbQMg2BCDLBL9eczWbIMyGr_6Jowjv2oDVOT3YbuH9LqZuydC0oGHj1oWoyqahKfEWzpA-SgRcTmAXHlOpaB_fNrSNGMuC5Hni0BLBKEWIdRAc0gwS4MqonGuPhD_CLB6OXjT_PAC-s96B2n-sLp2RAbY6syp_6sTQ2rTmYaf4pX3zpLVvffogkTgn7ghU43Mn76a1PAxVqwbxZeoUZSg5HPK7G416SKaETb015YlLY9U70bbu5m0mzLK1nt--kkgjQWVEfBtoFW_5JYL8-pKEdSGA4B1JRrK26GOLlbhivKXLyVvYbfw84TYfUUQQ79ZRmmRr3kWAmO1vzC063ljG1O-rv3_XOQ8WC-Cyw0koq77T9gvmRFFeDdYHNCs-rl2CQopZAzuCb4pS8o09cY6Xcsd3yWvL7qtVA_7Mce9mdJCUVMaPjdaa5szNvLDJ1VolvYNfJOt6Dh2NRxArU7AsV-8C0n4_CxVMsCjIrk23baYEQd9GOsE4zlz9c3ICdF_vRIc8T0aBr2xotSGdbbVG6wzA1FvwQifXG8jOX5av3rd3k-NXj07w8HEAPGtdkg4I_iFsJTnd9MRJbX5e010Q7QI875MJuHXr-A6nEiMIkpS_vEH0WOt_WUw","encrypted_key":"cFsAds8n7OkX71nAjY6hv7s_WmaFtT4-imMzW_Hzlgip27kAGXvbfg","header":{"epk":{"crv":"P-256","kty":"EC","x":"65XKDRaIyrljgOpouh09KQhWGXJTX27B85zD6-ju8kQ","y":"Cj2dg-BQQo1X_BPoQujdjlrBiEPRj15z0_xHLr_r6Ww"}},"iv":"FDzH68dzflzHj8RB","protected":"eyJhbGciOiJFQ0RILUVTK0EyNTZLVyIsImVuYyI6IkEyNTZHQ00iLCJzaGFyZWRIZWFkZXIiOnsicGF5bG9hZCI6ImV5SnphR0Z5WlVsa0lqb2lNREkxT0dVek5EWXRabVV5WlMwME1qVmpMV0ZrWXpJdE1XTTFaV000WmpObE16WmpJaXdpYjNkdVpYSWlPaUp5WldObGFYWmxjaUlzSW5KbFkyVnBkbVZ5Y3lJNld5SnlaV05sYVhabGNpSmRmUSIsInByb3RlY3RlZCI6ImV5SmhiR2NpT2lKRlV6STFOaUo5Iiwic2lnbmF0dXJlIjoiZm1GVjhLckJPV1djQzZ0Q1pMTVRUQjNEV0ZydXJLdk1JUTd1bWN6MTI0UFY3cjJ6Y0tuRXlpYUtJVlFDVEdCa250UThDYXJib3U0MjdFblB4aFYwSncifX0","tag":"kKU2-GYlnw9iQh8d33nOpA"}';

pythonDecryptAB = '{"ciphertext":"CzBFfVHH223qfPTmYAVTHW167o-zyI9yez8wL7b_aU803qdEZxwNWF7gh6p3NJDZL4YoKAOvO9nT6mYYPMNRy_Ef4plvtlbCa271-epr6STK9pxoaSazNaWIAevgVrRHrXtaxekENrajh2Zu8pDcP0ag8PZNjWFtWWcEyF4XQsmbcC6TqUs0Jfq-ylLSla8LJdJ_UBuvnHqQ4cd5Jt_Ix6Xqr1enCSEzjvHYp84W4JydY1spTGVlqJMVJ3o_AmxCPGFXd9M7FcKWoaq4ENGmDkVzmfqdDLPOEZzep8AZ0xbJ4kLEGLBSUZU3YT95btfMs8OKuRFSQ5TKc3kF4WGKRQgOT5KN3-XthLQ5zYpYVjrDw7oW5sIEKsea-TbUdFVfpj-JLCrh9ouCUlRROblbTo16y0xj69DTOIYcAsEoqOv2ElyeW4XArxSjADWj5DFYEPZG8bXmLqyf6OV3swmwSt4B34TJT1jGYVaTQapIOI4VL41JxuOvClq8RRnICg3-8R0mylNwumvhtqR21Z3jYJtYuZtZMLhwRuw2EsCdlMF5wvCI3ogmeDOFBUbF_-DSRQ-03stj3KHLRkYlzZySMRSzukQrjFyyrtpKAixbe02fotRXgQUUweTDxVcY7P04PdCzp4-gl7BFSjgT28t8h310NNotCNUE1YJSewqtqEL1uCpMoDCfbN_YHy3kWP1RrWV7_jNUPxGqYtZ77V3D_wfcrfkX2sbANj7-heEDrJeSMc3Te18ghRGEUCG1JfK8Rx1nDdrzzP07vCHGTox9DB4xrGPEcTzA6GfyR7Vsew0C24mPR3gFhKj-n2HkgEpdjIZrGhzDzqOGBW335mxh-39hTeBSMxvdWC4G8wvuczbeHlq-1VffpYMR-CfgfW2Kf_y_fIyO-DAC6WdRJe3NAQJ7uuqnvTvKkKG_fgBcnXqJ2RKh0sspOjYtOFoJyG8mDELo3Gt3cbOk","iv":"acbTS1uAJbcdJSUW","protected":"eyJhbGciOiJFQ0RILUVTK0EyNTZLVyIsImVuYyI6IkEyNTZHQ00iLCJzaGFyZWRIZWFkZXIiOnsicGF5bG9hZCI6ImV5SnphR0Z5WlVsa0lqb2lOREkzTlRjeU5qTXRNemhrTmkwME1qY3dMVGxqT1dJdE1HWm1NV1l4WlRjME1XUmtJaXdpYjNkdVpYSWlPaUp5WldObGFYWmxjaUlzSW5KbFkyVnBkbVZ5Y3lJNld5SnlaV05sYVhabGNpSXNJbk5sYm1SbGNpSmRmUSIsInByb3RlY3RlZCI6ImV5SmhiR2NpT2lKRlV6STFOaUo5Iiwic2lnbmF0dXJlIjoicUU4YllDOFI4SzBPbnJndGNKdXNiRFpsZEFoOEVmMVpockNQSEVLRFlyaUUxV2hJQm0tNmRrSlpGdlVzNGoxRzk5eWthTXBZbWpxS3JjTzctcXdMNncifX0","recipients":[{"encrypted_key":"WtIEYRlcGm-f2mvHtVVfd6kvoas8JoCHct588BehjR-pzrxF1WPsjA","header":{"epk":{"crv":"P-256","kty":"EC","x":"uSBaUaWBEFzNzCxIsYSK7wxquCLNbF-mZTJ0A3gefsU","y":"SdIArwd5q5Bpw6wZolAPVwAWXKN3wcm3xNla4hP09rA"}}},{"encrypted_key":"IeLDWoWw970dtDmo1KeKJtHbm90-9e5sd5FhgF6yTFgxrU1uKM2brQ","header":{"epk":{"crv":"P-256","kty":"EC","x":"m7NjzJqYeORHj3ynG39yNgfeCzr2vMJ88uilCynlwSs","y":"eddAA468pyeeShgBj_k5nqMVJiHrnsXu84LUET9brLM"}}}],"tag":"wJ_9nBmqrvHS7-WjCl4Gqg"}';

jsDecryptB = '{"ciphertext":"RPWoayrSjO1q57jh8WcJzFquQUmD4f8shujAVqpsMourphsGOGJJmPD9sXQLg8vne1RHxYOClf7cG5mlwj-MuvpthpmKCTlWY6V2_bkQEALh_IZuvm61a0mr3gtQtt7Mu_BHBKEHZOSQi50Sn_Y-n6HEp92V9ZFiqSh-sAP1wwgzJA4G2_0aAxYJnjILUvtLHk7pnz81sizCvH2IqCNJOKTqDCcqt3RHTiM08djLGe6C7Z577mPRM3JW-LOMbn3jlJRHpv5erLDJpbcewB7WcWQeEgQvg77U72pCo6BA6p_puNgrgBlyvs-UkT4A3PmKd46VAEP4yy-1u9jZeWX9ZfwAZIYPFkWI7I6_C0xyX5uwudkQlOk9HzwwVsMK1ysMHHsMe-Q0fXCoQngjcl7nOh9F35HwDNaIWqy09lovksqPR7PVzdeoxQlT1UfqSxDY9cxdtdZHvVlDTjDVrSG6v_dcmuM9FNLsRwJTgzhD4qrFPxiiZNOKSNBob2wDn9Aui-so1bdcVEGD-KGQABgZitKGnmQdjh7SEtbLUvxuvuf-CXZJ0Bwvs5fSa6GP6xtm8_qaCrc6awgPh8-CscWNkTWtte2XGD1T92J_MOx3QDKyyJw8BFVNZEFRNA2H0EjwhS3eZDizl2PDGP3FABaRcgBAazGTincpvXiB5RLTw0155ADAsD2fKQWSMudHe6WeUtf1HHKr0eqGyLUrYZZMIOpqAoj8aq3U3IS0tsf4UxK-RNzQd40pA5FEYGSFbLqYtc2_nvOt0_JgnuOrBAcWFBIAdDUcUrAfka1OhxCqTfHCPx5OQ36TlftV1cj5pQpc2Z1DPxe8AbbZ8wsxNnYDKlAQbEZTdHE0kA2a8kKLOPJJcxqTqUQfqkdOvDLuHAHdoWQSdxfiMaTqaFmso34-UJmQRq3W3d6NdghpRwLPW_hbN5F0BE4KUALUyfa2iSwUA1Dm4qR35WXaj9V7","iv":"7GwrPjDWr3mWkfUU","recipients":[{"encrypted_key":"teBliE1mhLH4MqU8ehT1LL6bDF6VjI32qKvMMzd3lRWDqTfp3nQGww","header":{"alg":"ECDH-ES+A256KW"}}],"tag":"WGmUG4H7wmGLghYqJDq3eg","protected":"eyJlbmMiOiJBMjU2R0NNIiwic2hhcmVkSGVhZGVyIjp7InNpZ25hdHVyZSI6Im9OemhIZFAwc2VjTDV3Y3VfbGxkeFkzVFM2b1FWX0Q0V3gyaXR6YjBjWkF3WmptbXNKcm1jbTA4RHZTNHBjb2ZobGUzR1F6R0JhUXFSS3E5UjQxWEpRIiwicGF5bG9hZCI6ImV5SnphR0Z5WlVsa0lqb2lPRFU0WkRNMk9HWXROR1F6TmkwMFpEZ3hMV0ZpWVdRdFlqWTFZVEl4WmpjNFpHTTFJaXdpYjNkdVpYSWlPaUp5WldObGFYWmxjaUlzSW5KbFkyVnBkbVZ5Y3lJNld5SnlaV05sYVhabGNpSmRmUSIsInByb3RlY3RlZCI6ImV5SmhiR2NpT2lKRlV6STFOaUo5In0sImVwayI6eyJ4IjoiX1Rva3JhYXFHS0w5dGIzZHVqTTFIVXY5Tm12WDNTN0RYTnNMVzhrUXhOWSIsImNydiI6IlAtMjU2Iiwia3R5IjoiRUMiLCJ5IjoidTdmcV9zbGREcUtHd3VsS0tieWNmOEFSa1NvTUVVVURiX18zbjlPdXZqSSJ9fQ"}';

jsDecryptAB = '{"ciphertext":"5L2bfeuZIp_fJvzJ0i31GPRkwogk93Gy6_rYbvGShLo6ouZhJe7ClwFMv8dzrcDsVbEuPR08wO1qgOpKWwdTEdq1mNfrWW7RpXWgEu52VPk3WXGdKrVt39QCnQlhMfVzlSd04H42_d3mu7AGL2RVDUoj3HNl8MIuLAQSPAC9dGeDv4NT73x1gEGViTWNOcATiILhlH9aihtYw_EBEEAtBE-LPMPwGoRnhay68JD3_0T_Yll0cPQT_coXx3I499uAuSdS_aJo4NOf90mVYRB-CMyWAFlD5I-0uhJyY0iazmbfYPipS2zBhpHoKhHvagZSUEaM2Y2tvpbfMO3lfs2yLFgJPa60KBTBIZdGaZAfhG1HLkgGiB1L_0qY5xI7QGSkrLgG23cRsbJnqifOXXe_gccG5fn3oaEMdpNT2LzHogOiLIdglSASHaE1CaZjMXvBzQ1-xOkCRmZkJzpy_rbwJAxdBjGR7nIbudXPkLpXVPSolQAyBvVQ37LsQ49aeEGpuyHM-bDb9XNKYMEPJlw0AQ-5Vg9Vd85_bUDNd6xKDB5N1Yr8hyk_TGK6_qj7Sm0VMW9N73U3nuyUb_0ilHGlDs-ra4HkV7-aXmymLiYqYeZkvbaeDr2oHlsVJ00MIb89SqejJL8CnEhWShb_5zCjFgwhRkhl4O4JYYkbcxyJLkv49fkAKGYZehkdrF5nL0iov-zgzX5HPufCUwsBJSCsHzg9hHy_oml8bvSxI_3t3pCLX3CYxhiomq12YFf_W6rWn0o1MS4YVArduXzvtU38tizYW3Wp2djOvT2EtHXj3-yQYFLB2zg3zQNQwCLpbbuIUkDjB1MjXjdFNhoxyS0PcjJDkuX1k3v8hPlSoE9tDg_EpvdYQtqTL4zHbN_rsvTb3b4S90TEzUW3cJ68IZACh4-Fo7edU5cxO3fnRtXb0E22QQy9ImGcsN6FkOpgME71VL_f5BufahY8qWrIAxY","iv":"PN7doBaqovb9pfln","recipients":[{"encrypted_key":"OXxxUKSlAAs70QuHGt73tNJAmhcEZ-VqdfkbuT2HVrd0iCAk5wGxcA","header":{"alg":"ECDH-ES+A256KW","epk":{"x":"dqoLrlbeMXTWqUJ_EB7Zxy5nN1QLrcMr3lgPJCXjDHE","crv":"P-256","kty":"EC","y":"hjsuKi_ftpeUWER831fVQPPqmwVfP7R7be-6HhIbLto"}}},{"encrypted_key":"sYQEkII22yh-jrz9lEKa1D3nsuBK_AF6B_9Ng7E94JTzaxIey2_UAQ","header":{"alg":"ECDH-ES+A256KW","epk":{"x":"jQlyTL4wUw5JjbpyboVH3_wZvYGGyT8ePx5kCplGJNc","crv":"P-256","kty":"EC","y":"DrstDm_exJuLJ8Vbpjqu4_8R8Y4ul5c7zzuh-kKUp7g"}}}],"tag":"Wd43r_immp33ZLhIQkuKJg","protected":"eyJlbmMiOiJBMjU2R0NNIiwic2hhcmVkSGVhZGVyIjp7InNpZ25hdHVyZSI6InVXb3dUbGVCMHhHTzFIVlloMURuSzdOVVFpSFhKb2dBM1lKQnVZc0FzbVhyaUNER0RKcnQ0Z0NqMlkzeGpCTmZXX0xjMlpsUG9naGtBcFRlaVo1U1FRIiwicGF5bG9hZCI6ImV5SnphR0Z5WlVsa0lqb2lNakZtTVRaaFpUQXRObVEzTmkwMFl6Um1MV0UxT1dJdE1tSXpOV1V4T1RaaFpXSmpJaXdpYjNkdVpYSWlPaUp5WldObGFYWmxjaUlzSW5KbFkyVnBkbVZ5Y3lJNld5SnlaV05sYVhabGNpSXNJbk5sYm1SbGNpSmRmUSIsInByb3RlY3RlZCI6ImV5SmhiR2NpT2lKRlV6STFOaUo5In19"}';


sender = UserManagement.importAuthenticatedUser('sender', pub_A, pub_A, priv_A, priv_A)
receiver = UserManagement.importAuthenticatedUser('receiver', pub_B, pub_B, priv_B, priv_B)


class TestCompatibilityGo(TestCase):
    """Test compatibility with go-it-crypto lib"""

    def test_single_receiver(self):
        """Test if receiver can decrypt goDecryptB"""
        log = receiver.decrypt(goDecryptB, create_fetch_sender([sender]))
        accessLog = log.extract()
        self.assertEqual(accessLog.justification, "go-it-crypto")

    def test_multiple_receiver(self):
        """Test if receiver and sender can decrypt goDecryptB"""
        log = receiver.decrypt(goDecryptAB, create_fetch_sender([sender, receiver]))
        accessLog = log.extract()
        self.assertEqual(accessLog.justification, "go-it-crypto")

        log = sender.decrypt(goDecryptAB, create_fetch_sender([sender, receiver]))
        accessLog = log.extract()
        self.assertEqual(accessLog.justification, "go-it-crypto")


class TestCompatibilityJS(TestCase):
    """Test compatibility with js-it-crypto lib"""

    def test_single_receiver(self):
        """Test if receiver can decrypt jsDecryptB"""
        log = receiver.decrypt(jsDecryptB, create_fetch_sender([sender]))
        accessLog = log.extract()
        self.assertEqual(accessLog.justification, "js-it-crypto")

    def test_multiple_receiver(self):
        """Test if receiver and sender can decrypt jsDecryptB"""
        log = receiver.decrypt(jsDecryptAB, create_fetch_sender([sender, receiver]))
        accessLog = log.extract()
        self.assertEqual(accessLog.justification, "js-it-crypto")

        log = sender.decrypt(jsDecryptAB, create_fetch_sender([sender, receiver]))
        accessLog = log.extract()
        self.assertEqual(accessLog.justification, "js-it-crypto")


class TestCompatibilityPython(TestCase):
    """Test compatibility with py-it-crypto lib"""

    def test_single_receiver(self):
        """Test if receiver can decrypt pyDecryptB"""
        log = receiver.decrypt(pythonDecryptB, create_fetch_sender([sender]))
        accessLog = log.extract()
        self.assertEqual(accessLog.justification, "py-it-crypto")

    def test_multiple_receiver(self):
        """Test if receiver and sender can decrypt pyDecryptB"""
        log = receiver.decrypt(pythonDecryptAB, create_fetch_sender([sender, receiver]))
        accessLog = log.extract()
        self.assertEqual(accessLog.justification, "py-it-crypto")

        log = sender.decrypt(pythonDecryptAB, create_fetch_sender([sender, receiver]))
        accessLog = log.extract()
        self.assertEqual(accessLog.justification, "py-it-crypto")


import json

from jwcrypto import jwk
from jwcrypto.common import json_decode

from logs.access_log import AccessLog
from user.authenticatedUser import AuthenticatedUser
from user.user import UserManagement


def proof_of_concept():
    invalid_ca_pem = """-----BEGIN CERTIFICATE-----
MIIBIDCByAIJAOGzO/GXoxxnMAoGCCqGSM49BAMCMBkxFzAVBgNVBAMMDkRldmVs
b3BtZW50IENBMB4XDTIyMTAyMDEyNTM1M1oXDTIzMTAyMDEyNTM1M1owGTEXMBUG
A1UEAwwORGV2ZWxvcG1lbnQgQ0EwWTATBgcqhkjOPQIBBggqhkjOPQMBBwNCAAQ6
/g/D3megZw6gssZIltWsk9CmqlqBNutzLjriJmTOGODHwHrTfdWIz8O161QB46jQ
qEgQrqzZK5H7X77BOlKwMAoGCCqGSM49BAMCA0cAMEQCICNvrFhNWpvO6vJAYGup
KiKEtPpfv6Rxe/Psq2XYy+H2AiA7fQHzny5CFJn4WsDDJGsgVOlnSD3gfLJ63uqq
M3s6nA==
-----END CERTIFICATE-----"""

    ca_pem = '-----BEGIN CERTIFICATE-----\n' + \
             'MIIBITCByAIJAJTQXJMDfhh5MAoGCCqGSM49BAMCMBkxFzAVBgNVBAMMDkRldmVs\n' + \
             'b3BtZW50IENBMB4XDTIyMTAxMDE1MzUzM1oXDTIzMTAxMDE1MzUzM1owGTEXMBUG\n' + \
             'A1UEAwwORGV2ZWxvcG1lbnQgQ0EwWTATBgcqhkjOPQIBBggqhkjOPQMBBwNCAAR0\n' + \
             'aTZBEZFtalbSmc8tNjh2UED6s09U4ZNM3fEA7AAOawH6RgQ1LjDtTFSAi0pO9YH4\n' + \
             'SVinZn6m4OwhGaoNZt0sMAoGCCqGSM49BAMCA0gAMEUCIQDtK9bAkAQHrAKmGPfV\n' + \
             'vg87jEqogKq85/q5V6jHZjawhwIgRUKldOc4fTa5/diT1OHKXLUW8uaDjZVNgv8Z\n' + \
             'HRVyXPs=\n' + \
             '-----END CERTIFICATE-----'

    keyA_pub = '-----BEGIN CERTIFICATE-----\n' + \
               'MIIBIDCByQIJAOuo8ugAq2wUMAkGByqGSM49BAEwGTEXMBUGA1UEAwwORGV2ZWxv\n' + \
               'cG1lbnQgQ0EwHhcNMjIxMDEwMTUzNTMzWhcNMjMxMDEwMTUzNTMzWjAbMRkwFwYD\n' + \
               'VQQDDBAibW1AZXhhbXBsZS5jb20iMFkwEwYHKoZIzj0CAQYIKoZIzj0DAQcDQgAE\n' + \
               'YlFye+p72EZ2z9xeBO9JAttfa/dhD6IhS6YpL1OixTkwiNA7CRU/tvGwlgdkVJPh\n' + \
               'QLhKldBRk37co8zLv3naszAJBgcqhkjOPQQBA0cAMEQCIDnDoDAmt4x7SSWVmYEs\n' + \
               '+JwLesjmZTkw0KaiZa+2E6ocAiBzPKTBADCCWDCGbiJg4V/7KV1tSiOYC9EpFOrk\n' + \
               'kyxIiA==\n' + \
               '-----END CERTIFICATE-----\n'

    keyA_priv = '-----BEGIN PRIVATE KEY-----\n' + \
                'MIGHAgEAMBMGByqGSM49AgEGCCqGSM49AwEHBG0wawIBAQQgAfMysADImEAjdKcY\n' + \
                '2sAIulabkZDyLdShbh+etB+RlZShRANCAARiUXJ76nvYRnbP3F4E70kC219r92EP\n' + \
                'oiFLpikvU6LFOTCI0DsJFT+28bCWB2RUk+FAuEqV0FGTftyjzMu/edqz\n' + \
                '-----END PRIVATE KEY-----'

    keyB_pub = '-----BEGIN CERTIFICATE-----\n' + \
               'MIIBITCByQIJAOuo8ugAq2wVMAkGByqGSM49BAEwGTEXMBUGA1UEAwwORGV2ZWxv\n' + \
               'cG1lbnQgQ0EwHhcNMjIxMDEwMTUzNTMzWhcNMjMxMDEwMTUzNTMzWjAbMRkwFwYD\n' + \
               'VQQDDBAibW1AZXhhbXBsZS5jb20iMFkwEwYHKoZIzj0CAQYIKoZIzj0DAQcDQgAE\n' + \
               'ELWdCySVeYt89xdfnUfbAh79CXk/gFvU8U988UpSLEAGx30aJ0ZecVpdKhlXO1G4\n' + \
               'yiyL8Sl6dypeN8iH7g3EtTAJBgcqhkjOPQQBA0gAMEUCIQCFDtrX9Mog3KA904Yp\n' + \
               'XduiWCtxVbGYGkSviklavTsNnAIgI8h9WNqHZdPJDVyhPwwS5oggTkGZah0LYfc3\n' + \
               '8qphvbY=\n' + \
               '-----END CERTIFICATE-----'

    keyB_priv = '-----BEGIN PRIVATE KEY-----\n' + \
                'MIGHAgEAMBMGByqGSM49AgEGCCqGSM49AwEHBG0wawIBAQQg9XQgYCk62PfcaOKE\n' + \
                'OlAerYQAx0EWg4eVfqMc1amEu0ehRANCAAQQtZ0LJJV5i3z3F1+dR9sCHv0JeT+A\n' + \
                'W9TxT3zxSlIsQAbHfRonRl5xWl0qGVc7UbjKLIvxKXp3Kl43yIfuDcS1\n' + \
                '-----END PRIVATE KEY-----'

    sender = UserManagement.importAuthenticatedUser("sender", keyA_pub, keyA_pub, keyA_priv,
                                                    keyA_priv)
    receiver = UserManagement.importAuthenticatedUser("receiver", keyB_pub, keyB_pub, keyB_priv,
                                                      keyB_priv)
    log_in = AccessLog.generate()
    log_in.monitor = sender.id
    log_in.owner = receiver.id
    log_in.justification = "py-it-crypto"
    singed_log = sender.sign_access_log(log_in)
    jwe = receiver.encrypt(singed_log, [receiver, sender])

    print(jwe)

    def fetchUser(id: str):
        return receiver

    signed_log = receiver.decrypt(jwe, fetchUser)
    log_out = signed_log.extract()
    print(log_in)
    print(log_out)
    # print(json.dumps(json.loads(jwe), indent=4))

    # root_cert = load_certificate(FILETYPE_PEM, ca_pem.encode())
    # untrusted_cert = load_certificate(FILETYPE_PEM, keyA_pub.encode())
    # store = X509Store()
    # store.add_cert(root_cert)
    # store_ctx = X509StoreContext(store, untrusted_cert)
    # print(store_ctx.verify_certificate())
    # return
    #
    # ca_key = load_pem_public_key(ca_pem.encode())
    # cert_to_check = x509.load_pem_x509_certificate(keyA_pub.encode())




proof_of_concept()



# Copyright (C) 2016 JWCrypto Project Contributors - see LICENSE file

import os
import struct
from abc import ABCMeta, abstractmethod
from binascii import hexlify, unhexlify

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import constant_time, hashes, hmac
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric import utils as ec_utils
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.kdf.concatkdf import ConcatKDFHash
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.primitives.keywrap import aes_key_unwrap, aes_key_wrap
from cryptography.hazmat.primitives.padding import PKCS7

from jwcrypto.common import InvalidCEKeyLength
from jwcrypto.common import InvalidJWAAlgorithm
from jwcrypto.common import InvalidJWEKeyLength
from jwcrypto.common import InvalidJWEKeyType
from jwcrypto.common import InvalidJWEOperation
from jwcrypto.common import base64url_decode, base64url_encode
from jwcrypto.common import json_decode
from jwcrypto.jwk import JWK

# Implements RFC 7518 - JSON Web Algorithms (JWA)


class JWAAlgorithm(metaclass=ABCMeta):

    @property
    @abstractmethod
    def name(self):
        """The algorithm Name"""

    @property
    @abstractmethod
    def description(self):
        """A short description"""

    @property
    @abstractmethod
    def keysize(self):
        """The actual/recommended/minimum key size"""

    @property
    @abstractmethod
    def algorithm_usage_location(self):
        """One of 'alg', 'enc' or 'JWK'"""

    @property
    @abstractmethod
    def algorithm_use(self):
        """One of 'sig', 'kex', 'enc'"""


def _bitsize(x):
    return len(x) * 8


def _inbytes(x):
    return x // 8


def _randombits(x):
    if x % 8 != 0:
        raise ValueError("length must be a multiple of 8")
    return os.urandom(_inbytes(x))


# Note: the number of bits should be a multiple of 16
def _encode_int(n, bits):
    e = '{:x}'.format(n)
    ilen = ((bits + 7) // 8) * 2  # number of bytes rounded up times 2 bytes
    return unhexlify(e.rjust(ilen, '0')[:ilen])


def _decode_int(n):
    return int(hexlify(n), 16)


class _RawJWS:

    def sign(self, key, payload):
        raise NotImplementedError

    def verify(self, key, payload, signature):
        raise NotImplementedError


class _RawHMAC(_RawJWS):

    def __init__(self, hashfn):
        self.backend = default_backend()
        self.hashfn = hashfn

    def _hmac_setup(self, key, payload):
        h = hmac.HMAC(key, self.hashfn, backend=self.backend)
        h.update(payload)
        return h

    def sign(self, key, payload):
        skey = base64url_decode(key.get_op_key('sign'))
        h = self._hmac_setup(skey, payload)
        return h.finalize()

    def verify(self, key, payload, signature):
        vkey = base64url_decode(key.get_op_key('verify'))
        h = self._hmac_setup(vkey, payload)
        h.verify(signature)


class _RawRSA(_RawJWS):
    def __init__(self, padfn, hashfn):
        self.padfn = padfn
        self.hashfn = hashfn

    def sign(self, key, payload):
        skey = key.get_op_key('sign')
        return skey.sign(payload, self.padfn, self.hashfn)

    def verify(self, key, payload, signature):
        pkey = key.get_op_key('verify')
        pkey.verify(signature, payload, self.padfn, self.hashfn)


class _RawEC(_RawJWS):
    def __init__(self, curve, hashfn):
        self._curve = curve
        self.hashfn = hashfn

    @property
    def curve(self):
        return self._curve

    def sign(self, key, payload):
        skey = key.get_op_key('sign', self._curve)
        size = skey.key_size
        signature = skey.sign(payload, ec.ECDSA(self.hashfn))
        r, s = ec_utils.decode_dss_signature(signature)
        return _encode_int(r, size) + _encode_int(s, size)

    def verify(self, key, payload, signature):
        pkey = key.get_op_key('verify', self._curve)
        r = signature[:len(signature) // 2]
        s = signature[len(signature) // 2:]
        enc_signature = ec_utils.encode_dss_signature(
            int(hexlify(r), 16), int(hexlify(s), 16))
        pkey.verify(enc_signature, payload, ec.ECDSA(self.hashfn))


class _RawNone(_RawJWS):

    def sign(self, key, payload):
        return ''

    def verify(self, key, payload, signature):
        if key['kty'] != 'oct' or key.get_op_key() != '':
            raise InvalidSignature('The "none" signature cannot be verified')


class _HS256(_RawHMAC, JWAAlgorithm):

    name = "HS256"
    description = "HMAC using SHA-256"
    keysize = 256
    algorithm_usage_location = 'alg'
    algorithm_use = 'sig'

    def __init__(self):
        super(_HS256, self).__init__(hashes.SHA256())


class _HS384(_RawHMAC, JWAAlgorithm):

    name = "HS384"
    description = "HMAC using SHA-384"
    keysize = 384
    algorithm_usage_location = 'alg'
    algorithm_use = 'sig'

    def __init__(self):
        super(_HS384, self).__init__(hashes.SHA384())


class _HS512(_RawHMAC, JWAAlgorithm):

    name = "HS512"
    description = "HMAC using SHA-512"
    keysize = 512
    algorithm_usage_location = 'alg'
    algorithm_use = 'sig'

    def __init__(self):
        super(_HS512, self).__init__(hashes.SHA512())


class _RS256(_RawRSA, JWAAlgorithm):

    name = "RS256"
    description = "RSASSA-PKCS1-v1_5 using SHA-256"
    keysize = 2048
    algorithm_usage_location = 'alg'
    algorithm_use = 'sig'

    def __init__(self):
        super(_RS256, self).__init__(padding.PKCS1v15(), hashes.SHA256())


class _RS384(_RawRSA, JWAAlgorithm):

    name = "RS384"
    description = "RSASSA-PKCS1-v1_5 using SHA-384"
    keysize = 2048
    algorithm_usage_location = 'alg'
    algorithm_use = 'sig'

    def __init__(self):
        super(_RS384, self).__init__(padding.PKCS1v15(), hashes.SHA384())


class _RS512(_RawRSA, JWAAlgorithm):

    name = "RS512"
    description = "RSASSA-PKCS1-v1_5 using SHA-512"
    keysize = 2048
    algorithm_usage_location = 'alg'
    algorithm_use = 'sig'

    def __init__(self):
        super(_RS512, self).__init__(padding.PKCS1v15(), hashes.SHA512())


class _ES256(_RawEC, JWAAlgorithm):

    name = "ES256"
    description = "ECDSA using P-256 and SHA-256"
    keysize = 256
    algorithm_usage_location = 'alg'
    algorithm_use = 'sig'

    def __init__(self):
        super(_ES256, self).__init__('P-256', hashes.SHA256())


class _ES256K(_RawEC, JWAAlgorithm):

    name = "ES256K"
    description = "ECDSA using secp256k1 curve and SHA-256"
    keysize = 256
    algorithm_usage_location = 'alg'
    algorithm_use = 'sig'

    def __init__(self):
        super(_ES256K, self).__init__('secp256k1', hashes.SHA256())


class _ES384(_RawEC, JWAAlgorithm):

    name = "ES384"
    description = "ECDSA using P-384 and SHA-384"
    keysize = 384
    algorithm_usage_location = 'alg'
    algorithm_use = 'sig'

    def __init__(self):
        super(_ES384, self).__init__('P-384', hashes.SHA384())


class _ES512(_RawEC, JWAAlgorithm):

    name = "ES512"
    description = "ECDSA using P-521 and SHA-512"
    keysize = 512
    algorithm_usage_location = 'alg'
    algorithm_use = 'sig'

    def __init__(self):
        super(_ES512, self).__init__('P-521', hashes.SHA512())


class _PS256(_RawRSA, JWAAlgorithm):

    name = "PS256"
    description = "RSASSA-PSS using SHA-256 and MGF1 with SHA-256"
    keysize = 2048
    algorithm_usage_location = 'alg'
    algorithm_use = 'sig'

    def __init__(self):
        padfn = padding.PSS(padding.MGF1(hashes.SHA256()),
                            hashes.SHA256.digest_size)
        super(_PS256, self).__init__(padfn, hashes.SHA256())


class _PS384(_RawRSA, JWAAlgorithm):

    name = "PS384"
    description = "RSASSA-PSS using SHA-384 and MGF1 with SHA-384"
    keysize = 2048
    algorithm_usage_location = 'alg'
    algorithm_use = 'sig'

    def __init__(self):
        padfn = padding.PSS(padding.MGF1(hashes.SHA384()),
                            hashes.SHA384.digest_size)
        super(_PS384, self).__init__(padfn, hashes.SHA384())


class _PS512(_RawRSA, JWAAlgorithm):

    name = "PS512"
    description = "RSASSA-PSS using SHA-512 and MGF1 with SHA-512"
    keysize = 2048
    algorithm_usage_location = 'alg'
    algorithm_use = 'sig'

    def __init__(self):
        padfn = padding.PSS(padding.MGF1(hashes.SHA512()),
                            hashes.SHA512.digest_size)
        super(_PS512, self).__init__(padfn, hashes.SHA512())


class _None(_RawNone, JWAAlgorithm):

    name = "none"
    description = "No digital signature or MAC performed"
    keysize = 0
    algorithm_usage_location = 'alg'
    algorithm_use = 'sig'


class _RawKeyMgmt:

    def wrap(self, key, bitsize, cek, headers):
        raise NotImplementedError

    def unwrap(self, key, bitsize, ek, headers):
        raise NotImplementedError


class _RSA(_RawKeyMgmt):

    def __init__(self, padfn):
        self.padfn = padfn

    def _check_key(self, key):
        if not isinstance(key, JWK):
            raise ValueError('key is not a JWK object')
        if key['kty'] != 'RSA':
            raise InvalidJWEKeyType('RSA', key['kty'])

    # FIXME: get key size and ensure > 2048 bits
    def wrap(self, key, bitsize, cek, headers):
        self._check_key(key)
        if not cek:
            cek = _randombits(bitsize)
        rk = key.get_op_key('wrapKey')
        ek = rk.encrypt(cek, self.padfn)
        return {'cek': cek, 'ek': ek}

    def unwrap(self, key, bitsize, ek, headers):
        self._check_key(key)
        rk = key.get_op_key('unwrapKey')
        cek = rk.decrypt(ek, self.padfn)
        if _bitsize(cek) != bitsize:
            raise InvalidJWEKeyLength(bitsize, _bitsize(cek))
        return cek


class _Rsa15(_RSA, JWAAlgorithm):

    name = 'RSA1_5'
    description = "RSAES-PKCS1-v1_5"
    keysize = 2048
    algorithm_usage_location = 'alg'
    algorithm_use = 'kex'

    def __init__(self):
        super(_Rsa15, self).__init__(padding.PKCS1v15())

    def unwrap(self, key, bitsize, ek, headers):
        self._check_key(key)
        # Address MMA attack by implementing RFC 3218 - 2.3.2. Random Filling
        # provides a random cek that will cause the decryption engine to
        # run to the end, but will fail decryption later.

        # always generate a random cek so we spend roughly the
        # same time as in the exception side of the branch
        cek = _randombits(bitsize)
        try:
            cek = super(_Rsa15, self).unwrap(key, bitsize, ek, headers)
            # always raise so we always run through the exception handling
            # code in all cases
            raise Exception('Dummy')
        except Exception:  # pylint: disable=broad-except
            return cek


class _RsaOaep(_RSA, JWAAlgorithm):

    name = 'RSA-OAEP'
    description = "RSAES OAEP using default parameters"
    keysize = 2048
    algorithm_usage_location = 'alg'
    algorithm_use = 'kex'

    def __init__(self):
        super(_RsaOaep, self).__init__(
            padding.OAEP(padding.MGF1(hashes.SHA1()),
                         hashes.SHA1(), None))


class _RsaOaep256(_RSA, JWAAlgorithm):  # noqa: ignore=N801

    name = 'RSA-OAEP-256'
    description = "RSAES OAEP using SHA-256 and MGF1 with SHA-256"
    keysize = 2048
    algorithm_usage_location = 'alg'
    algorithm_use = 'kex'

    def __init__(self):
        super(_RsaOaep256, self).__init__(
            padding.OAEP(padding.MGF1(hashes.SHA256()),
                         hashes.SHA256(), None))


class _AesKw(_RawKeyMgmt):

    keysize = None

    def __init__(self):
        self.backend = default_backend()

    def _get_key(self, key, op):
        if not isinstance(key, JWK):
            raise ValueError('key is not a JWK object')
        if key['kty'] != 'oct':
            raise InvalidJWEKeyType('oct', key['kty'])
        rk = base64url_decode(key.get_op_key(op))
        if _bitsize(rk) != self.keysize:
            raise InvalidJWEKeyLength(self.keysize, _bitsize(rk))
        return rk

    def wrap(self, key, bitsize, cek, headers):
        rk = self._get_key(key, 'encrypt')
        if not cek:
            cek = _randombits(bitsize)

        ek = aes_key_wrap(rk, cek, default_backend())

        return {'cek': cek, 'ek': ek}

    def unwrap(self, key, bitsize, ek, headers):
        rk = self._get_key(key, 'decrypt')

        cek = aes_key_unwrap(rk, ek, default_backend())
        if _bitsize(cek) != bitsize:
            raise InvalidJWEKeyLength(bitsize, _bitsize(cek))
        return cek


class _A128KW(_AesKw, JWAAlgorithm):

    name = 'A128KW'
    description = "AES Key Wrap using 128-bit key"
    keysize = 128
    algorithm_usage_location = 'alg'
    algorithm_use = 'kex'


class _A192KW(_AesKw, JWAAlgorithm):

    name = 'A192KW'
    description = "AES Key Wrap using 192-bit key"
    keysize = 192
    algorithm_usage_location = 'alg'
    algorithm_use = 'kex'


class _A256KW(_AesKw, JWAAlgorithm):

    name = 'A256KW'
    description = "AES Key Wrap using 256-bit key"
    keysize = 256
    algorithm_usage_location = 'alg'
    algorithm_use = 'kex'


class _AesGcmKw(_RawKeyMgmt):

    keysize = None

    def __init__(self):
        self.backend = default_backend()

    def _get_key(self, key, op):
        if not isinstance(key, JWK):
            raise ValueError('key is not a JWK object')
        if key['kty'] != 'oct':
            raise InvalidJWEKeyType('oct', key['kty'])
        rk = base64url_decode(key.get_op_key(op))
        if _bitsize(rk) != self.keysize:
            raise InvalidJWEKeyLength(self.keysize, _bitsize(rk))
        return rk

    def wrap(self, key, bitsize, cek, headers):
        rk = self._get_key(key, 'encrypt')
        if not cek:
            cek = _randombits(bitsize)

        iv = _randombits(96)
        cipher = Cipher(algorithms.AES(rk), modes.GCM(iv),
                        backend=self.backend)
        encryptor = cipher.encryptor()
        ek = encryptor.update(cek) + encryptor.finalize()

        tag = encryptor.tag
        return {'cek': cek, 'ek': ek,
                'header': {'iv': base64url_encode(iv),
                           'tag': base64url_encode(tag)}}

    def unwrap(self, key, bitsize, ek, headers):
        rk = self._get_key(key, 'decrypt')

        if 'iv' not in headers:
            raise ValueError('Invalid Header, missing "iv" parameter')
        iv = base64url_decode(headers['iv'])
        if 'tag' not in headers:
            raise ValueError('Invalid Header, missing "tag" parameter')
        tag = base64url_decode(headers['tag'])

        cipher = Cipher(algorithms.AES(rk), modes.GCM(iv, tag),
                        backend=self.backend)
        decryptor = cipher.decryptor()
        cek = decryptor.update(ek) + decryptor.finalize()
        if _bitsize(cek) != bitsize:
            raise InvalidJWEKeyLength(bitsize, _bitsize(cek))
        return cek


class _A128GcmKw(_AesGcmKw, JWAAlgorithm):

    name = 'A128GCMKW'
    description = "Key wrapping with AES GCM using 128-bit key"
    keysize = 128
    algorithm_usage_location = 'alg'
    algorithm_use = 'kex'


class _A192GcmKw(_AesGcmKw, JWAAlgorithm):

    name = 'A192GCMKW'
    description = "Key wrapping with AES GCM using 192-bit key"
    keysize = 192
    algorithm_usage_location = 'alg'
    algorithm_use = 'kex'


class _A256GcmKw(_AesGcmKw, JWAAlgorithm):

    name = 'A256GCMKW'
    description = "Key wrapping with AES GCM using 256-bit key"
    keysize = 256
    algorithm_usage_location = 'alg'
    algorithm_use = 'kex'


class _Pbes2HsAesKw(_RawKeyMgmt):

    name = None
    keysize = None
    hashsize = None

    def __init__(self):
        self.backend = default_backend()
        self.aeskwmap = {128: _A128KW, 192: _A192KW, 256: _A256KW}

    def _get_key(self, alg, key, p2s, p2c):
        if not isinstance(key, JWK):
            # backwards compatibility for old interface
            if isinstance(key, bytes):
                plain = key
            else:
                plain = key.encode('utf8')
        else:
            plain = base64url_decode(key.get_op_key())

        salt = bytes(self.name.encode('utf8')) + b'\x00' + p2s

        if self.hashsize == 256:
            hashalg = hashes.SHA256()
        elif self.hashsize == 384:
            hashalg = hashes.SHA384()
        elif self.hashsize == 512:
            hashalg = hashes.SHA512()
        else:
            raise ValueError('Unknown Hash Size')

        kdf = PBKDF2HMAC(algorithm=hashalg, length=_inbytes(self.keysize),
                         salt=salt, iterations=p2c, backend=self.backend)
        rk = kdf.derive(plain)
        if _bitsize(rk) != self.keysize:
            raise InvalidJWEKeyLength(self.keysize, len(rk))
        return JWK(kty="oct", use="enc", k=base64url_encode(rk))

    def wrap(self, key, bitsize, cek, headers):
        ret_header = {}
        if 'p2s' in headers:
            p2s = base64url_decode(headers['p2s'])
            if len(p2s) < 8:
                raise ValueError('Invalid Salt, must be 8 or more octects')
        else:
            p2s = _randombits(128)
            ret_header['p2s'] = base64url_encode(p2s)
        if 'p2c' in headers:
            p2c = headers['p2c']
        else:
            p2c = 8192
            ret_header['p2c'] = p2c
        kek = self._get_key(headers['alg'], key, p2s, p2c)

        aeskw = self.aeskwmap[self.keysize]()
        ret = aeskw.wrap(kek, bitsize, cek, headers)
        if len(ret_header) > 0:
            ret['header'] = ret_header
        return ret

    def unwrap(self, key, bitsize, ek, headers):
        if 'p2s' not in headers:
            raise ValueError('Invalid Header, missing "p2s" parameter')
        if 'p2c' not in headers:
            raise ValueError('Invalid Header, missing "p2c" parameter')
        p2s = base64url_decode(headers['p2s'])
        p2c = headers['p2c']
        kek = self._get_key(headers['alg'], key, p2s, p2c)

        aeskw = self.aeskwmap[self.keysize]()
        return aeskw.unwrap(kek, bitsize, ek, headers)


class _Pbes2Hs256A128Kw(_Pbes2HsAesKw, JWAAlgorithm):

    name = 'PBES2-HS256+A128KW'
    description = 'PBES2 with HMAC SHA-256 and "A128KW" wrapping'
    keysize = 128
    algorithm_usage_location = 'alg'
    algorithm_use = 'kex'
    hashsize = 256


class _Pbes2Hs384A192Kw(_Pbes2HsAesKw, JWAAlgorithm):

    name = 'PBES2-HS384+A192KW'
    description = 'PBES2 with HMAC SHA-384 and "A192KW" wrapping'
    keysize = 192
    algorithm_usage_location = 'alg'
    algorithm_use = 'kex'
    hashsize = 384


class _Pbes2Hs512A256Kw(_Pbes2HsAesKw, JWAAlgorithm):

    name = 'PBES2-HS512+A256KW'
    description = 'PBES2 with HMAC SHA-512 and "A256KW" wrapping'
    keysize = 256
    algorithm_usage_location = 'alg'
    algorithm_use = 'kex'
    hashsize = 512


class _Direct(_RawKeyMgmt, JWAAlgorithm):

    name = 'dir'
    description = "Direct use of a shared symmetric key"
    keysize = 128
    algorithm_usage_location = 'alg'
    algorithm_use = 'kex'

    def _check_key(self, key):
        if not isinstance(key, JWK):
            raise ValueError('key is not a JWK object')
        if key['kty'] != 'oct':
            raise InvalidJWEKeyType('oct', key['kty'])

    def wrap(self, key, bitsize, cek, headers):
        self._check_key(key)
        if cek:
            return (cek, None)
        k = base64url_decode(key.get_op_key('encrypt'))
        if _bitsize(k) != bitsize:
            raise InvalidCEKeyLength(bitsize, _bitsize(k))
        return {'cek': k}

    def unwrap(self, key, bitsize, ek, headers):
        self._check_key(key)
        if ek != b'':
            raise ValueError('Invalid Encryption Key.')
        cek = base64url_decode(key.get_op_key('decrypt'))
        if _bitsize(cek) != bitsize:
            raise InvalidJWEKeyLength(bitsize, _bitsize(cek))
        return cek


class _EcdhEs(_RawKeyMgmt, JWAAlgorithm):

    name = 'ECDH-ES'
    description = "ECDH-ES using Concat KDF"
    algorithm_usage_location = 'alg'
    algorithm_use = 'kex'
    keysize = None

    def __init__(self):
        self.backend = default_backend()
        self.aeskwmap = {128: _A128KW, 192: _A192KW, 256: _A256KW}

    def _check_key(self, key):
        if not isinstance(key, JWK):
            raise ValueError('key is not a JWK object')
        if key['kty'] not in ['EC', 'OKP']:
            raise InvalidJWEKeyType('EC or OKP', key['kty'])
        if key['kty'] == 'OKP':
            if key['crv'] not in ['X25519', 'X448']:
                raise InvalidJWEKeyType('X25519 or X448',
                                        key['crv'])

    def _derive(self, privkey, pubkey, alg, bitsize, headers):
        # OtherInfo is defined in NIST SP 56A 5.8.1.2.1

        # AlgorithmID
        otherinfo = struct.pack('>I', len(alg))
        otherinfo += bytes(alg.encode('utf8'))

        # PartyUInfo
        apu = base64url_decode(headers['apu']) if 'apu' in headers else b''
        otherinfo += struct.pack('>I', len(apu))
        otherinfo += apu

        # PartyVInfo
        apv = base64url_decode(headers['apv']) if 'apv' in headers else b''
        otherinfo += struct.pack('>I', len(apv))
        otherinfo += apv

        # SuppPubInfo
        otherinfo += struct.pack('>I', bitsize)

        # no SuppPrivInfo

        # Shared Key generation
        if isinstance(privkey, ec.EllipticCurvePrivateKey):
            shared_key = privkey.exchange(ec.ECDH(), pubkey)
        else:
            # X25519/X448
            shared_key = privkey.exchange(pubkey)

        ckdf = ConcatKDFHash(algorithm=hashes.SHA256(),
                             length=_inbytes(bitsize),
                             otherinfo=otherinfo,
                             backend=self.backend)
        return ckdf.derive(shared_key)

    def wrap(self, key, bitsize, cek, headers):
        self._check_key(key)
        dk_size = self.keysize
        if self.keysize is None:
            if cek is not None:
                raise InvalidJWEOperation('ECDH-ES cannot use an existing CEK')
            alg = headers['enc']
            dk_size = bitsize
        else:
            alg = headers['alg']

        epk = JWK.generate(kty=key['kty'], crv=key['crv'])
        dk = self._derive(epk.get_op_key('unwrapKey'),
                          key.get_op_key('wrapKey'),
                          alg, dk_size, headers)

        if self.keysize is None:
            ret = {'cek': dk}
        else:
            aeskw = self.aeskwmap[self.keysize]()
            kek = JWK(kty="oct", use="enc", k=base64url_encode(dk))
            ret = aeskw.wrap(kek, bitsize, cek, headers)

        ret['header'] = {'epk': json_decode(epk.export_public())}
        return ret

    def unwrap(self, key, bitsize, ek, headers):
        if 'epk' not in headers:
            raise ValueError('Invalid Header, missing "epk" parameter')
        self._check_key(key)
        dk_size = self.keysize
        if self.keysize is None:
            alg = headers['enc']
            dk_size = bitsize
        else:
            alg = headers['alg']

        epk = JWK(**headers['epk'])
        dk = self._derive(key.get_op_key('unwrapKey'),
                          epk.get_op_key('wrapKey'),
                          alg, dk_size, headers)
        if self.keysize is None:
            return dk
        else:
            aeskw = self.aeskwmap[self.keysize]()
            kek = JWK(kty="oct", use="enc", k=base64url_encode(dk))
            cek = aeskw.unwrap(kek, bitsize, ek, headers)
            return cek


class _EcdhEsAes128Kw(_EcdhEs):

    name = 'ECDH-ES+A128KW'
    description = 'ECDH-ES using Concat KDF and "A128KW" wrapping'
    keysize = 128
    algorithm_usage_location = 'alg'
    algorithm_use = 'kex'


class _EcdhEsAes192Kw(_EcdhEs):

    name = 'ECDH-ES+A192KW'
    description = 'ECDH-ES using Concat KDF and "A192KW" wrapping'
    keysize = 192
    algorithm_usage_location = 'alg'
    algorithm_use = 'kex'


class _EcdhEsAes256Kw(_EcdhEs):

    name = 'ECDH-ES+A256KW'
    description = 'ECDH-ES using Concat KDF and "A256KW" wrapping'
    keysize = 256
    algorithm_usage_location = 'alg'
    algorithm_use = 'kex'


class _EdDsa(_RawJWS, JWAAlgorithm):

    name = 'EdDSA'
    description = 'EdDSA using Ed25519 or Ed448 algorithms'
    algorithm_usage_location = 'alg'
    algorithm_use = 'sig'
    keysize = None

    def sign(self, key, payload):
        if key['crv'] in ['Ed25519', 'Ed448']:
            skey = key.get_op_key('sign')
            return skey.sign(payload)
        raise NotImplementedError

    def verify(self, key, payload, signature):
        if key['crv'] in ['Ed25519', 'Ed448']:
            pkey = key.get_op_key('verify')
            return pkey.verify(signature, payload)
        raise NotImplementedError


class _RawJWE:

    def encrypt(self, k, a, m):
        raise NotImplementedError

    def decrypt(self, k, a, iv, e, t):
        raise NotImplementedError


class _AesCbcHmacSha2(_RawJWE):

    keysize = None

    def __init__(self, hashfn):
        self.backend = default_backend()
        self.hashfn = hashfn
        self.blocksize = algorithms.AES.block_size
        self.wrap_key_size = self.keysize * 2

    def _mac(self, k, a, iv, e):
        al = _encode_int(_bitsize(a), 64)
        h = hmac.HMAC(k, self.hashfn, backend=self.backend)
        h.update(a)
        h.update(iv)
        h.update(e)
        h.update(al)
        m = h.finalize()
        return m[:_inbytes(self.keysize)]

    # RFC 7518 - 5.2.2
    def encrypt(self, k, a, m):
        """ Encrypt according to the selected encryption and hashing
        functions.

        :param k: Encryption key (optional)
        :param a: Additional Authentication Data
        :param m: Plaintext

        Returns a dictionary with the computed data.
        """
        hkey = k[:_inbytes(self.keysize)]
        ekey = k[_inbytes(self.keysize):]

        # encrypt
        iv = _randombits(self.blocksize)
        cipher = Cipher(algorithms.AES(ekey), modes.CBC(iv),
                        backend=self.backend)
        encryptor = cipher.encryptor()
        padder = PKCS7(self.blocksize).padder()
        padded_data = padder.update(m) + padder.finalize()
        e = encryptor.update(padded_data) + encryptor.finalize()

        # mac
        t = self._mac(hkey, a, iv, e)

        return (iv, e, t)

    def decrypt(self, k, a, iv, e, t):
        """ Decrypt according to the selected encryption and hashing
        functions.
        :param k: Encryption key (optional)
        :param a: Additional Authenticated Data
        :param iv: Initialization Vector
        :param e: Ciphertext
        :param t: Authentication Tag

        Returns plaintext or raises an error
        """
        hkey = k[:_inbytes(self.keysize)]
        dkey = k[_inbytes(self.keysize):]

        # verify mac
        if not constant_time.bytes_eq(t, self._mac(hkey, a, iv, e)):
            raise InvalidSignature('Failed to verify MAC')

        # decrypt
        cipher = Cipher(algorithms.AES(dkey), modes.CBC(iv),
                        backend=self.backend)
        decryptor = cipher.decryptor()
        d = decryptor.update(e) + decryptor.finalize()
        unpadder = PKCS7(self.blocksize).unpadder()
        return unpadder.update(d) + unpadder.finalize()


class _A128CbcHs256(_AesCbcHmacSha2, JWAAlgorithm):

    name = 'A128CBC-HS256'
    description = "AES_128_CBC_HMAC_SHA_256 authenticated"
    keysize = 128
    algorithm_usage_location = 'enc'
    algorithm_use = 'enc'

    def __init__(self):
        super(_A128CbcHs256, self).__init__(hashes.SHA256())


class _A192CbcHs384(_AesCbcHmacSha2, JWAAlgorithm):

    name = 'A192CBC-HS384'
    description = "AES_192_CBC_HMAC_SHA_384 authenticated"
    keysize = 192
    algorithm_usage_location = 'enc'
    algorithm_use = 'enc'

    def __init__(self):
        super(_A192CbcHs384, self).__init__(hashes.SHA384())


class _A256CbcHs512(_AesCbcHmacSha2, JWAAlgorithm):

    name = 'A256CBC-HS512'
    description = "AES_256_CBC_HMAC_SHA_512 authenticated"
    keysize = 256
    algorithm_usage_location = 'enc'
    algorithm_use = 'enc'

    def __init__(self):
        super(_A256CbcHs512, self).__init__(hashes.SHA512())


class _AesGcm(_RawJWE):

    keysize = None

    def __init__(self):
        self.backend = default_backend()
        self.wrap_key_size = self.keysize

    # RFC 7518 - 5.3
    def encrypt(self, k, a, m):
        """ Encrypt accoriding to the selected encryption and hashing
        functions.

        :param k: Encryption key (optional)
        :param a: Additional Authentication Data
        :param m: Plaintext

        Returns a dictionary with the computed data.
        """
        iv = _randombits(96)
        cipher = Cipher(algorithms.AES(k), modes.GCM(iv),
                        backend=self.backend)
        encryptor = cipher.encryptor()
        encryptor.authenticate_additional_data(a)
        e = encryptor.update(m) + encryptor.finalize()

        return (iv, e, encryptor.tag)

    def decrypt(self, k, a, iv, e, t):
        """ Decrypt accoriding to the selected encryption and hashing
        functions.
        :param k: Encryption key (optional)
        :param a: Additional Authenticated Data
        :param iv: Initialization Vector
        :param e: Ciphertext
        :param t: Authentication Tag

        Returns plaintext or raises an error
        """
        cipher = Cipher(algorithms.AES(k), modes.GCM(iv, t),
                        backend=self.backend)
        decryptor = cipher.decryptor()
        decryptor.authenticate_additional_data(a)
        return decryptor.update(e) + decryptor.finalize()


class _A128Gcm(_AesGcm, JWAAlgorithm):

    name = 'A128GCM'
    description = "AES GCM using 128-bit key"
    keysize = 128
    algorithm_usage_location = 'enc'
    algorithm_use = 'enc'


class _A192Gcm(_AesGcm, JWAAlgorithm):

    name = 'A192GCM'
    description = "AES GCM using 192-bit key"
    keysize = 192
    algorithm_usage_location = 'enc'
    algorithm_use = 'enc'


class _A256Gcm(_AesGcm, JWAAlgorithm):

    name = 'A256GCM'
    description = "AES GCM using 256-bit key"
    keysize = 256
    algorithm_usage_location = 'enc'
    algorithm_use = 'enc'


class JWA:
    """JWA Signing Algorithms.

    This class provides access to all JWA algorithms.
    """

    algorithms_registry = {
        'HS256': _HS256,
        'HS384': _HS384,
        'HS512': _HS512,
        'RS256': _RS256,
        'RS384': _RS384,
        'RS512': _RS512,
        'ES256': _ES256,
        'ES256K': _ES256K,
        'ES384': _ES384,
        'ES512': _ES512,
        'PS256': _PS256,
        'PS384': _PS384,
        'PS512': _PS512,
        'none': _None,
        'RSA1_5': _Rsa15,
        'RSA-OAEP': _RsaOaep,
        'RSA-OAEP-256': _RsaOaep256,
        'A128KW': _A128KW,
        'A192KW': _A192KW,
        'A256KW': _A256KW,
        'dir': _Direct,
        'ECDH-ES': _EcdhEs,
        'ECDH-ES+A128KW': _EcdhEsAes128Kw,
        'ECDH-ES+A192KW': _EcdhEsAes192Kw,
        'ECDH-ES+A256KW': _EcdhEsAes256Kw,
        'EdDSA': _EdDsa,
        'A128GCMKW': _A128GcmKw,
        'A192GCMKW': _A192GcmKw,
        'A256GCMKW': _A256GcmKw,
        'PBES2-HS256+A128KW': _Pbes2Hs256A128Kw,
        'PBES2-HS384+A192KW': _Pbes2Hs384A192Kw,
        'PBES2-HS512+A256KW': _Pbes2Hs512A256Kw,
        'A128CBC-HS256': _A128CbcHs256,
        'A192CBC-HS384': _A192CbcHs384,
        'A256CBC-HS512': _A256CbcHs512,
        'A128GCM': _A128Gcm,
        'A192GCM': _A192Gcm,
        'A256GCM': _A256Gcm
    }

    @classmethod
    def instantiate_alg(cls, name, use=None):
        alg = cls.algorithms_registry[name]
        if use is not None and alg.algorithm_use != use:
            raise KeyError
        return alg()

    @classmethod
    def signing_alg(cls, name):
        try:
            return cls.instantiate_alg(name, use='sig')
        except KeyError:
            raise InvalidJWAAlgorithm('%s is not a valid Signing algorithm'
                                      ' name' % name) from None

    @classmethod
    def keymgmt_alg(cls, name):
        try:
            return cls.instantiate_alg(name, use='kex')
        except KeyError:
            raise InvalidJWAAlgorithm('%s is not a valid Key Management'
                                      ' algorithm name' % name) from None

    @classmethod
    def encryption_alg(cls, name):
        try:
            return cls.instantiate_alg(name, use='enc')
        except KeyError:
            raise InvalidJWAAlgorithm('%s is not a valid Encryption'
                                      ' algorithm name' % name) from None



from typing import Callable

from jwcrypto import jws
from jwcrypto.common import json_encode
from jwcrypto.jwk import JWK

from py_it_crypto.crypto.decryption import DecryptionService
from py_it_crypto.crypto.encryption import EncryptionService
from py_it_crypto.globals import SIGNING_ALG
from py_it_crypto.logs.access_log import AccessLog, SignedAccessLog
from py_it_crypto.user.remoteUser import RemoteUser


class AuthenticatedUser(RemoteUser):
    decryption_key: JWK
    signing_key: JWK

    def __init__(self, id: str, encryption_certificate: JWK, verification_certificate: JWK,
                 decryption_key: JWK, signing_key: JWK):
        super().__init__(id, encryption_certificate, verification_certificate)
        self.decryption_key = decryption_key
        self.signing_key = signing_key

    def encrypt(self, log: SignedAccessLog, receivers: list[RemoteUser]) -> str:
        return EncryptionService.encrypt(jwsAccessLog=log, sender=self, receivers=receivers)

    def decrypt(self, jwe: str, fetch_user: Callable[[str], RemoteUser]) -> SignedAccessLog:
        return DecryptionService.decrypt(jwe=jwe, receiver=self, fetch_user=fetch_user)

    def sign_data(self, data: bytes) -> str:
        token = jws.JWS(data)
        token.add_signature(self.signing_key, None, json_encode({"alg": SIGNING_ALG}))
        return token.serialize()

    def sign_access_log(self, log: AccessLog) -> SignedAccessLog:
        singed = self.sign_data(log.to_bytes())
        return SignedAccessLog.from_json(singed)
